@@ -55,13 +55,14 @@ import sys
 import time
 import traceback as traceback_
 import logging
+import platform
 
 import six
 from six.moves import queue
 from six.moves import urllib
 
 from . import errors, __version__
-from ._compat import ntob, bton, ntou, unquote_to_bytes
+from ._compat import bton, ntou
 from .workers import threadpool
 from .makefile import MakeFile
 
@@ -71,7 +72,7 @@ __all__ = ('HTTPRequest', 'HTTPConnection', 'HTTPServer',
            'Gateway', 'get_ssl_adapter_class')
 
 
-if 'win' in sys.platform and hasattr(socket, 'AF_INET6'):
+if platform.system() == 'Windows' and hasattr(socket, 'AF_INET6'):
     if not hasattr(socket, 'IPPROTO_IPV6'):
         socket.IPPROTO_IPV6 = 41
     if not hasattr(socket, 'IPV6_V6ONLY'):
@@ -92,13 +93,12 @@ QUOTED_SLASH_REGEX = re.compile(b'(?i)' + QUOTED_SLASH)
 
 
 comma_separated_headers = [
-    ntob(h) for h in
-    ['Accept', 'Accept-Charset', 'Accept-Encoding',
-     'Accept-Language', 'Accept-Ranges', 'Allow', 'Cache-Control',
-     'Connection', 'Content-Encoding', 'Content-Language', 'Expect',
-     'If-Match', 'If-None-Match', 'Pragma', 'Proxy-Authenticate', 'TE',
-     'Trailer', 'Transfer-Encoding', 'Upgrade', 'Vary', 'Via', 'Warning',
-     'WWW-Authenticate']
+    b'Accept', b'Accept-Charset', b'Accept-Encoding',
+    b'Accept-Language', b'Accept-Ranges', b'Allow', b'Cache-Control',
+    b'Connection', b'Content-Encoding', b'Content-Language', b'Expect',
+    b'If-Match', b'If-None-Match', b'Pragma', b'Proxy-Authenticate', b'TE',
+    b'Trailer', b'Transfer-Encoding', b'Upgrade', b'Vary', b'Via', b'Warning',
+    b'WWW-Authenticate',
 ]
 
 
@@ -122,7 +122,8 @@ class HeaderReader(object):
         Headers which are repeated are folded together using a comma if their
         specification so dictates.
 
-        This function raises ValueError when the read bytes violate the HTTP spec.
+        This function raises ValueError when the read bytes violate the HTTP
+        spec.
         You should probably return "400 Bad Request" if this happens.
         """
         if hdict is None:
@@ -402,7 +403,8 @@ class ChunkedRFile(object):
         self.bytes_read += len(line)
 
         if self.maxlen and self.bytes_read > self.maxlen:
-            raise errors.MaxSizeExceeded('Request Entity Too Large', self.maxlen)
+            raise errors.MaxSizeExceeded(
+                'Request Entity Too Large', self.maxlen)
 
         line = line.strip().split(SEMICOLON, 1)
 
@@ -697,8 +699,18 @@ class HTTPRequest(object):
 
         try:
             method, uri, req_protocol = request_line.strip().split(SPACE, 2)
-            req_protocol_str = req_protocol.decode('ascii')
-            rp = int(req_protocol_str[5]), int(req_protocol_str[7])
+            if not req_protocol.startswith(b'HTTP/'):
+                self.simple_response(
+                    '400 Bad Request', 'Malformed Request-Line: bad protocol'
+                )
+                return False
+            rp = req_protocol[5:].split(b'.', 1)
+            rp = tuple(map(int, rp))  # Minor.Major must be threat as integers
+            if rp > (1, 1):
+                self.simple_response(
+                    '505 HTTP Version Not Supported', 'Cannot fulfill request'
+                )
+                return False
         except (ValueError, IndexError):
             self.simple_response('400 Bad Request', 'Malformed Request-Line')
             return False
@@ -707,45 +719,97 @@ class HTTPRequest(object):
         self.method = method.upper()
 
         if self.strict_mode and method != self.method:
-            self.simple_response(
-                '400 Bad Request',
-                'Malformed method name: According to RFC 2616 (section 5.1.1) and its successors '
-                'RFC 7230 (section 3.1.1) and RFC 7231 (section 4.1) method names are case-sensitive and uppercase.'
+            resp = (
+                'Malformed method name: According to RFC 2616 '
+                '(section 5.1.1) and its successors '
+                'RFC 7230 (section 3.1.1) and RFC 7231 (section 4.1) '
+                'method names are case-sensitive and uppercase.'
             )
+            self.simple_response('400 Bad Request', resp)
             return False
 
-        scheme = authority = path = qs = EMPTY
+        try:
+            if six.PY2:  # FIXME: Figure out better way to do this
+                # Ref: https://stackoverflow.com/a/196392/595220 (like this?)
+                """This is a dummy check for unicode in URI."""
+                ntou(bton(uri, 'ascii'), 'ascii')
+            scheme, authority, path, qs, fragment = urllib.parse.urlsplit(uri)
+        except UnicodeError:
+            self.simple_response('400 Bad Request', 'Malformed Request-URI')
+            return False
 
         if self.method == b'OPTIONS':
             # TODO: cover this branch with tests
             path = (uri
                     # https://tools.ietf.org/html/rfc7230#section-5.3.4
                     if self.proxy_mode or uri == ASTERISK
-                    else urllib.parse.urlsplit(uri).path)
+                    else path)
         elif self.method == b'CONNECT':
             # TODO: cover this branch with tests
             if not self.proxy_mode:
-                self.simple_response('400 Bad Request')
+                self.simple_response('405 Method Not Allowed')
                 return False
 
-            # https://tools.ietf.org/html/rfc7230#section-5.3.3
-            authority = uri
-        else:
+            # `urlsplit()` above parses "example.com:3128" as path part of URI.
+            # this is a workaround, which makes it detect netloc correctly
+            uri_split = urllib.parse.urlsplit(b'//' + uri)
+            _scheme, _authority, _path, _qs, _fragment = uri_split
+            _port = EMPTY
             try:
-                # https://tools.ietf.org/html/rfc7230#section-5.3.1 (origin_form) and
-                # https://tools.ietf.org/html/rfc7230#section-5.3.2 (absolute form)
-                if six.PY2:  # FIXME: Figure out better way to do this
-                    # Ref: https://stackoverflow.com/a/196392/595220 (like this?)
-                    """This is a dummy check for unicode in URI."""
-                    ntou(bton(uri, 'ascii'), 'ascii')
-                scheme, authority, path, qs, fragment = urllib.parse.urlsplit(uri)
-            except UnicodeError:
-                self.simple_response('400 Bad Request', 'Malformed Request-URI')
+                _port = uri_split.port
+            except ValueError:
+                pass
+
+            # FIXME: use third-party validation to make checks against RFC
+            # the validation doesn't take into account, that urllib parses
+            # invalid URIs without raising errors
+            # https://tools.ietf.org/html/rfc7230#section-5.3.3
+            invalid_path = (
+                _authority != uri
+                or not _port
+                or any((_scheme, _path, _qs, _fragment))
+            )
+            if invalid_path:
+                self.simple_response('400 Bad Request',
+                                     'Invalid path in Request-URI: request-'
+                                     'target must match authority-form.')
                 return False
 
-            if (self.strict_mode and not self.proxy_mode) and (scheme or authority):
-                self.simple_response('400 Bad Request',
-                                     'Absolute URI not allowed if server is not a proxy.')
+            authority = path = _authority
+            scheme = qs = fragment = EMPTY
+        else:
+            uri_is_absolute_form = (scheme or authority)
+
+            disallowed_absolute = (
+                self.strict_mode
+                and not self.proxy_mode
+                and uri_is_absolute_form
+            )
+            if disallowed_absolute:
+                # https://tools.ietf.org/html/rfc7230#section-5.3.2
+                # (absolute form)
+                """Absolute URI is only allowed within proxies."""
+                self.simple_response(
+                    '400 Bad Request',
+                    'Absolute URI not allowed if server is not a proxy.',
+                )
+                return False
+
+            invalid_path = (
+                self.strict_mode
+                and not uri.startswith(FORWARD_SLASH)
+                and not uri_is_absolute_form
+            )
+            if invalid_path:
+                # https://tools.ietf.org/html/rfc7230#section-5.3.1
+                # (origin_form) and
+                """Path should start with a forward slash."""
+                resp = (
+                    'Invalid path in Request-URI: request-target must contain '
+                    'origin-form which starts with absolute-path (URI '
+                    'starting with a slash "/").'
+                )
+                self.simple_response('400 Bad Request', resp)
                 return False
 
             if fragment:
@@ -765,12 +829,13 @@ class HTTPRequest(object):
             # But note that "...a URI must be separated into its components
             # before the escaped characters within those components can be
             # safely decoded." http://www.ietf.org/rfc/rfc2396.txt, sec 2.4.2
-            # Therefore, "/this%2Fpath" becomes "/this%2Fpath", not "/this/path".
+            # Therefore, "/this%2Fpath" becomes "/this%2Fpath", not
+            # "/this/path".
             try:
                 # TODO: Figure out whether exception can really happen here.
                 # It looks like it's caught on urlsplit() call above.
                 atoms = [
-                    unquote_to_bytes(x)
+                    urllib.parse.unquote_to_bytes(x)
                     for x in QUOTED_SLASH_REGEX.split(path)
                 ]
             except ValueError as ex:
@@ -907,10 +972,8 @@ class HTTPRequest(object):
             self.rfile = KnownLengthRFile(self.conn.rfile, cl)
 
         self.server.gateway(self).respond()
+        self.ready and self.ensure_headers_sent()
 
-        if (self.ready and not self.sent_headers):
-            self.sent_headers = True
-            self.send_headers()
         if self.chunked_write:
             self.conn.wfile.write(b'0\r\n\r\n')
 
@@ -951,6 +1014,12 @@ class HTTPRequest(object):
             if ex.args[0] not in errors.socket_errors_to_ignore:
                 raise
 
+    def ensure_headers_sent(self):
+        """Ensure headers are sent to the client if not already sent."""
+        if not self.sent_headers:
+            self.sent_headers = True
+            self.send_headers()
+
     def write(self, chunk):
         """Write unbuffered data to the client."""
         if self.chunked_write and chunk:
@@ -978,7 +1047,11 @@ class HTTPRequest(object):
             if status < 200 or status in (204, 205, 304):
                 pass
             else:
-                if self.response_protocol == 'HTTP/1.1' and self.method != b'HEAD':
+                needs_chunked = (
+                    self.response_protocol == 'HTTP/1.1'
+                    and self.method != b'HEAD'
+                )
+                if needs_chunked:
                     # Use the chunked transfer-coding
                     self.chunked_write = True
                     self.outheaders.append((b'Transfer-Encoding', b'chunked'))
@@ -1048,7 +1121,8 @@ class HTTPConnection(object):
 
         Args:
             server (HTTPServer): web server object receiving this request
-            socket (socket._socketobject): the raw socket object (usually TCP) for this connection
+            socket (socket._socketobject): the raw socket object (usually
+                TCP) for this connection
             makefile (file): a fileobject class for reading from the socket
         """
         self.server = server
@@ -1085,52 +1159,28 @@ class HTTPConnection(object):
         except socket.error as ex:
             errnum = ex.args[0]
             # sadly SSL sockets return a different (longer) time out string
-            if (
-                errnum == 'timed out' or
-                errnum == 'The read operation timed out'
-            ):
+            timeout_errs = 'timed out', 'The read operation timed out'
+            if errnum in timeout_errs:
                 # Don't error if we're between requests; only error
                 # if 1) no request has been started at all, or 2) we're
                 # in the middle of a request.
                 # See https://github.com/cherrypy/cherrypy/issues/853
                 if (not request_seen) or (req and req.started_request):
-                    # Don't bother writing the 408 if the response
-                    # has already started being written.
-                    if req and not req.sent_headers:
-                        try:
-                            req.simple_response('408 Request Timeout')
-                        except errors.FatalSSLAlert:
-                            # Close the connection.
-                            return
-                        except errors.NoSSLError:
-                            self._handle_no_ssl(req)
+                    self._conditional_error(req, '408 Request Timeout')
             elif errnum not in errors.socket_errors_to_ignore:
                 self.server.error_log('socket.error %s' % repr(errnum),
                                       level=logging.WARNING, traceback=True)
-                if req and not req.sent_headers:
-                    try:
-                        req.simple_response('500 Internal Server Error')
-                    except errors.FatalSSLAlert:
-                        # Close the connection.
-                        return
-                    except errors.NoSSLError:
-                        self._handle_no_ssl(req)
-            return
+                self._conditional_error(req, '500 Internal Server Error')
         except (KeyboardInterrupt, SystemExit):
             raise
         except errors.FatalSSLAlert:
-            # Close the connection.
-            return
+            pass
         except errors.NoSSLError:
             self._handle_no_ssl(req)
         except Exception as ex:
-            self.server.error_log(repr(ex), level=logging.ERROR, traceback=True)
-            if req and not req.sent_headers:
-                try:
-                    req.simple_response('500 Internal Server Error')
-                except errors.FatalSSLAlert:
-                    # Close the connection.
-                    return
+            self.server.error_log(
+                repr(ex), level=logging.ERROR, traceback=True)
+            self._conditional_error(req, '500 Internal Server Error')
 
     linger = False
 
@@ -1146,6 +1196,22 @@ class HTTPConnection(object):
         req.simple_response('400 Bad Request', msg)
         self.linger = True
 
+    def _conditional_error(self, req, response):
+        """Respond with an error.
+
+        Don't bother writing if a response
+        has already started being written.
+        """
+        if not req or req.sent_headers:
+            return
+
+        try:
+            req.simple_response(response)
+        except errors.FatalSSLAlert:
+            pass
+        except errors.NoSSLError:
+            self._handle_no_ssl(req)
+
     def close(self):
         """Close the socket underlying this connection."""
         self.rfile.close()
@@ -1220,7 +1286,9 @@ class HTTPServer(object):
     """The minimum number of worker threads to create (default 10)."""
 
     maxthreads = None
-    """The maximum number of worker threads to create (default -1 = no limit)."""
+    """The maximum number of worker threads to create.
+
+    (default -1 = no limit)"""
 
     server_name = None
     """The name of the server; defaults to ``self.version``."""
@@ -1232,10 +1300,14 @@ class HTTPServer(object):
     features used in the response."""
 
     request_queue_size = 5
-    """The 'backlog' arg to socket.listen(); max queued connections (default 5)."""
+    """The 'backlog' arg to socket.listen(); max queued connections.
+
+    (default 5)."""
 
     shutdown_timeout = 5
-    """The total time, in seconds, to wait for worker threads to cleanly exit."""
+    """The total time to wait for worker threads to cleanly exit.
+
+    Specified in seconds."""
 
     timeout = 10
     """The timeout in seconds for accepted connections (default 10)."""
@@ -1250,7 +1322,7 @@ class HTTPServer(object):
     """
 
     ready = False
-    """An internal flag which marks whether the socket is accepting connections."""
+    """Internal flag which indicating the socket is accepting connections."""
 
     max_request_header_size = 0
     """The maximum size, in bytes, for request headers, or 0 for no limit."""
@@ -1270,7 +1342,9 @@ class HTTPServer(object):
     You must have the corresponding SSL driver library installed.
     """
 
-    def __init__(self, bind_addr, gateway, minthreads=10, maxthreads=-1, server_name=None):
+    def __init__(
+            self, bind_addr, gateway, minthreads=10, maxthreads=-1,
+            server_name=None):
         """Initialize HTTPServer instance.
 
         Args:
@@ -1278,22 +1352,17 @@ class HTTPServer(object):
             gateway (Gateway): gateway for processing HTTP requests
             minthreads (int): minimum number of threads for HTTP thread pool
             maxthreads (int): maximum number of threads for HTTP thread pool
-            server_name (str): web server name to be advertised via Server HTTP header
+            server_name (str): web server name to be advertised via Server
+                HTTP header
         """
         self.bind_addr = bind_addr
         self.gateway = gateway
 
-        self.requests = threadpool.ThreadPool(self, min=minthreads or 1, max=maxthreads)
+        self.requests = threadpool.ThreadPool(
+            self, min=minthreads or 1, max=maxthreads)
 
         if not server_name:
             server_name = self.version
-            """
-            Default server name. It is used to fall back to ``socket.gethostname()``.
-            Ref: cherrypy/cheroot#33. If the previous logic needed the following should restore it:
-
-            >>> from cheroot._compat import import ntou, tonative
-            >>> ntou(tonative(ntou(tonative(hn, 'utf-8'), 'utf-8').encode('idna')))
-            """
         self.server_name = server_name
         self.clear_stats()
 
@@ -1342,10 +1411,27 @@ class HTTPServer(object):
         return '%s.%s(%r)' % (self.__module__, self.__class__.__name__,
                               self.bind_addr)
 
-    def _get_bind_addr(self):
+    @property
+    def bind_addr(self):
+        """Return the interface on which to listen for connections.
+
+        For TCP sockets, a (host, port) tuple. Host values may be any IPv4
+        or IPv6 address, or any valid hostname. The string 'localhost' is a
+        synonym for '127.0.0.1' (or '::1', if your hosts file prefers IPv6).
+        The string '0.0.0.0' is a special IPv4 entry meaning "any active
+        interface" (INADDR_ANY), and '::' is the similar IN6ADDR_ANY for
+        IPv6. The empty string or None are not allowed.
+
+        For UNIX sockets, supply the filename as a string.
+
+        Systemd socket activation is automatic and doesn't require tempering
+        with this variable.
+        """
         return self._bind_addr
 
-    def _set_bind_addr(self, value):
+    @bind_addr.setter
+    def bind_addr(self, value):
+        """Set the interface on which to listen for connections."""
         if isinstance(value, tuple) and value[0] in ('', None):
             # Despite the socket module docs, using '' does not
             # allow AI_PASSIVE to work. Passing None instead
@@ -1361,22 +1447,6 @@ class HTTPServer(object):
                              "Use '0.0.0.0' (IPv4) or '::' (IPv6) instead "
                              'to listen on all active interfaces.')
         self._bind_addr = value
-    bind_addr = property(
-        _get_bind_addr,
-        _set_bind_addr,
-        doc="""The interface on which to listen for connections.
-
-        For TCP sockets, a (host, port) tuple. Host values may be any IPv4
-        or IPv6 address, or any valid hostname. The string 'localhost' is a
-        synonym for '127.0.0.1' (or '::1', if your hosts file prefers IPv6).
-        The string '0.0.0.0' is a special IPv4 entry meaning "any active
-        interface" (INADDR_ANY), and '::' is the similar IN6ADDR_ANY for
-        IPv6. The empty string or None are not allowed.
-
-        For UNIX sockets, supply the filename as a string.
-
-        Systemd socket activation is automatic and doesn't require tempering
-        with this variable""")
 
     def safe_start(self):
         """Run the server forever, and stop it cleanly on exit."""
@@ -1415,13 +1485,13 @@ class HTTPServer(object):
             # So we can reuse the socket...
             try:
                 os.unlink(self.bind_addr)
-            except:
+            except Exception:
                 pass
 
             # So everyone can access the socket...
             try:
                 os.chmod(self.bind_addr, 0o777)
-            except:
+            except Exception:
                 pass
 
             info = [
@@ -1436,12 +1506,14 @@ class HTTPServer(object):
                     host, port, socket.AF_UNSPEC,
                     socket.SOCK_STREAM, 0, socket.AI_PASSIVE)
             except socket.gaierror:
-                if ':' in self.bind_addr[0]:
-                    info = [(socket.AF_INET6, socket.SOCK_STREAM,
-                             0, '', self.bind_addr + (0, 0))]
-                else:
-                    info = [(socket.AF_INET, socket.SOCK_STREAM,
-                             0, '', self.bind_addr)]
+                sock_type = socket.AF_INET
+                bind_addr = self.bind_addr
+
+                if ':' in host:
+                    sock_type = socket.AF_INET6
+                    bind_addr = bind_addr + (0, 0)
+
+                info = [(sock_type, socket.SOCK_STREAM, 0, '', bind_addr)]
 
         if not self.socket:
             msg = 'No socket could be created'
@@ -1473,7 +1545,7 @@ class HTTPServer(object):
                 self.tick()
             except (KeyboardInterrupt, SystemExit):
                 raise
-            except:
+            except Exception:
                 self.error_log('Error in HTTPServer.tick', level=logging.ERROR,
                                traceback=True)
 
@@ -1504,7 +1576,11 @@ class HTTPServer(object):
         """Create (or recreate) the actual socket object."""
         self.socket = socket.socket(family, type, proto)
         prevent_socket_inheritance(self.socket)
-        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
+        if platform.system() != 'Windows':
+            # Windows has different semantics for SO_REUSEADDR,
+            # so don't set it.
+            # https://msdn.microsoft.com/en-us/library/ms740621(v=vs.85).aspx
+            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
         if self.nodelay and not isinstance(self.bind_addr, str):
             self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
 
@@ -1516,7 +1592,12 @@ class HTTPServer(object):
         # If listening on the IPV6 any address ('::' = IN6ADDR_ANY),
         # activate dual-stack. See
         # https://github.com/cherrypy/cherrypy/issues/871.
-        if hasattr(socket, 'AF_INET6') and family == socket.AF_INET6 and host in ('::', '::0', '::0.0.0.0'):
+        listening_ipv6 = (
+            hasattr(socket, 'AF_INET6')
+            and family == socket.AF_INET6
+            and host in ('::', '::0', '::0.0.0.0')
+        )
+        if listening_ipv6:
             try:
                 self.socket.setsockopt(
                     socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
@@ -1526,6 +1607,19 @@ class HTTPServer(object):
                 pass
 
         self.socket.bind(self.bind_addr)
+        # TODO: keep requested bind_addr separate real bound_addr (port is
+        # different in case of ephemeral port 0)
+        self.bind_addr = self.socket.getsockname()
+        if family in (
+            # Windows doesn't have socket.AF_UNIX, so not using it in check
+            socket.AF_INET,
+            socket.AF_INET6,
+        ):
+            """UNIX domain sockets are strings or bytes.
+
+            In case of bytes with a leading null-byte it's an abstract socket.
+            """
+            self.bind_addr = self.bind_addr[:2]
 
     def tick(self):
         """Accept a new connection and put it on the Queue."""
@@ -1619,16 +1713,17 @@ class HTTPServer(object):
                 return
             raise
 
-    def _get_interrupt(self):
+    @property
+    def interrupt(self):
+        """Flag interrupt of the server."""
         return self._interrupt
 
-    def _set_interrupt(self, interrupt):
+    @interrupt.setter
+    def interrupt(self, interrupt):
+        """Perform the shutdown of this server and save the exception."""
         self._interrupt = True
         self.stop()
         self._interrupt = interrupt
-    interrupt = property(_get_interrupt, _set_interrupt,
-                         doc='Set this to an Exception instance to '
-                             'interrupt the server.')
 
     def stop(self):
         """Gracefully shutdown a server that is serving forever."""
@@ -1677,7 +1772,7 @@ class HTTPServer(object):
 
 
 class Gateway(object):
-    """A base class to interface HTTPServer with other systems, such as WSGI."""
+    """Base class to interface HTTPServer with other systems, such as WSGI."""
 
     def __init__(self, req):
         """Initialize Gateway instance with request.
@@ -1689,7 +1784,7 @@ class Gateway(object):
 
     def respond(self):
         """Process the current request. Must be overridden in a subclass."""
-        raise NotImplemented
+        raise NotImplementedError
 
 
 # These may either be ssl.Adapter subclasses or the string names
